// File for ulsop compatibility (ULVM)
// Includes functions for exectuing UXE-Files and an interrupthandler
// For Ulsop-functions

// Note: some calls are slightly different to ulsop (ULDIF-Tag)

UCHR iretdata = 0xCF;
UINT dstr[32];
long globl;
unsigned globa,globb,globc,globd;
char *globfar;

ulkbhit(int pid)
    {
    int i;
    for(i=0;i<MSGTAB_MAX;i++)
        if(tasks[pid].msg[i].used)
            if(tasks[pid].msg[i].msgid == MSG_KEYPRESS)
                return 1;
    return 0;
    }

ulgetch(int pid)
    {
    int i;
    for(i=0;i<MSGTAB_MAX;i++)
        if(tasks[pid].msg[i].used)
            if(tasks[pid].msg[i].msgid == MSG_KEYPRESS)
                {
                tasks[pid].msg[i].used = 0;
                return tasks[pid].msg[i].b;
                }
    // No msg? wait
    while(1)
        {
        i = getmessage(pid);    // Phew, executing ul-apps is so HARD, take a break
        if(tasks[pid].msg[i].msgid == MSG_KEYPRESS)
            return tasks[pid].msg[i].b;
        }
    }

ulgetwnd(int pid)
    {
    int i;
    if(!uldat) return;
    for(i=0;i<ULWM_MAXWIN;i++)
        if(uldat->windat[i].owner == pid)
            return i;
    return 0;
    }

ulgetsig(int pid, int clear)
    {
    int i;
    // To get a signal we have to parse the msg. queue and get relevant stuff
    for(i=0;i<MSGTAB_MAX;i++)
        {
//        printf("msg used:%i, ID=%i",tasks[pid].msg[i].used,tasks[pid].msg[i].msgid);
        if(tasks[pid].msg[i].used)
            {
            if(tasks[pid].msg[i].msgid == MSG_KILL || tasks[pid].msg[i].msgid == MSG_CLOSEWND)
                {
                if(clear) tasks[pid].msg[i].used = 0;
                return 2; 
                }
            if(tasks[pid].msg[i].msgid == MSG_PAINT)
                {
                printf("got paint-message");
                if(clear) tasks[pid].msg[i].used = 0;
                return 1; 
                }
            else if(tasks[pid].msg[i].msgid != MSG_KEYPRESS)
                tasks[pid].msg[i].used = 0; // This message is ulsop-incompatible, remove it
            }
        }
    return 0;
    }

// void __far __interrupt ulhandle()
void ulhandle()
    {
    UINT rax,rbx,rcx,rdx,rds,res,rsi,rdi,curt;
    int i,j;
    _asm
        {
        mov rax, ax
        mov rbx, bx
        mov rcx, cx
        mov rdx, dx
        mov rsi, si
        mov rdi, di
        mov ax, ds
        mov rds, ax
        mov ax, es
        mov res, ax
        mov ax, SEG dummy
        mov ds, ax
        sti
        }
    printf("ULAPI: %X",dstr[0]);
    switch(dstr[0])
        {
        case 0x0:       // Kill task (ULDIF: No need to pass cs in reg 1)
            task_quit(curtask);
            break;
        case 0x2:       // Get foreground task-id (ULDIF: Ohps, Ulsop allowes
                        // only 1 wnd/app - this is incompatible, assume the
                        // running task is in fg
            // sameasgetpid            
        case 0x1:       // Get Pid
            dstr[1] = curtask;
            break;
        case 0x3:       // Get last signal & clear
            if(dstr[1] == 0xFF)
                dstr[1] = curtask;
            if(dstr[1] == 0xFF)
                dstr[2] = 0xFF;
            else
                {
                dstr[2] = ulgetsig(dstr[1],1);
                printf("got signal %i from %i",dstr[2],dstr[1]);
                }
            break;
        // Function 4 unsupported
        case 0x5:       // Function 5: Get task-data (1=id ; 2,3,4,5,6,7 = staus,segname,offname,lastsig,ss,sp)
                if(dstr[1] == 0xFF)
                    dstr[1] = curtask;
                dstr[2] = tasks[dstr[1]].used ? 0 : -1;
                dstr[3] = FP_SEG(tasks[dstr[1]].filename);  // ULDIF: Filename instead of taskname
                dstr[4] = FP_OFF(tasks[dstr[1]].filename);
                dstr[5] = ulgetsig(dstr[1],0);
                dstr[6] = tasks[dstr[1]].ss;
                dstr[7] = tasks[dstr[1]].sp;
                break;
        case 0x6:       // Function 6: undocumented
                idle();     // UlDIF: Function 6 used by ULMULTI.COM, here is it just an idle-call
                break;
        case 0x9:       // Function 9: kbhit?
                dstr[1] = ulkbhit(curtask);
                break;
        case 0xA:       // Function a: getch
                dstr[1] = ulgetch(curtask);
                break;
        case 0xB:       // Function B: Get setting (1,2=name ; 3,4=value)
                dstr[3] = 0;    // ULDIF: No INI-Services TODO: Add a replacement
                dstr[4] = 0;
                break;
        case 0xC:       // Function C: Set setting (1,2,3,4 = name,value ; 5: (=0: OK))
        case 0xD:       // Function D: Save settings to disk (; 1: (=0:OK))
                dstr[1] = 0;    // ULDIF: No INI-Services
                dstr[5] = 0;    
                break;
        case 0xE:       // Function E: Int 2 string (1,2,3,4 = num,buff,len)
                itoa(dstr[1],MK_FP(dstr[2],dstr[3]),dstr[4]);
                break;
        case 0xF:       // Function F: Execute (1,2=filename)
                task_create(MK_FP(dstr[1],dstr[2]),"",curtask);
                break;
        case 0x10:      // Function 10: Get timer 1/100 sec. (; 1,2 = low,high)
                globl = clock();
                dstr[1] = globl;
                dstr[2] = globl / 65536;
                break;
        case 0x11:      // Function 11: Randomize
                dstr[2] = (UINT) ((( (long) rand()) * dstr[1]) / RAND_MAX);
                break;
        case 0x12:      // Function 12: Call DOS-Prog (1,2 = seg,off)
                system((char *) MK_FP(dstr[1],dstr[2]));
                break;
        case 0x7B:      // Function 7B: OK-Messagebox (1,2 = seg,off)
                ok_msgbox(MK_FP(dstr[1],dstr[2]));
                break;
        case 0x7C:      // Function 7C: Get rect for splitted screen (6,7,8,9 = x-piece,y-piece,x-size,y-size; 2,3,4,5 = x1,y1,x2,y2)
                dstr[2] = dstr[6] * dstr[8];
                dstr[3] = dstr[7] * dstr[9];
                dstr[4] = dstr[2] + dstr[8] - 1;
                dstr[5] = dstr[3] + dstr[9] - 1;
                break;
        case 0x7D:      // Function 7D: Split screen into pieces (1,2,3,4 = x-pieces,y-pieces,x-reserve,y-reserve ; 5,6 = x-size,y-size)
                dstr[5] = (curx - dstr[3])                       / dstr[1];
                dstr[6] = (cury - tbheight - hbheight - dstr[4]) / dstr[2];
                break;
        case 0x7E:      // Function 7E: Get center-pos
                dstr[2] = (curx / 2) - (dstr[1] / 2);
                break;
        case 0x7F:      // Function 7F: Get available resolution (1,2,3 = x,y,c)
                dstr[1] = curx;
                dstr[2] = cury - tbheight - hbheight;
                dstr[3] = 0;
                break;
        case 0x80:      // Function 80: Get resolution (1,2,3 = x,y,c)
                dstr[1] = curx;
                dstr[2] = cury;
                dstr[3] = 0;
                break;
        case 0x81:      // Function 81: Set resolution (1,2,3 = x,y,c)
                break;  // ULDIF: Unsupported
        case 0x82:      // Function 82: Draw something (9 can be 0x1234 to force drawing)
                if(dstr[9] == 0x1234) j = 0, dstr[9] = 0; else j = ulgetwnd(curtask);
                printf("82h subfunction %.4X, wnd %i",dstr[1],j);
                switch(dstr[1])
                        {
                        case 0:         // Function 82/0: Set pixel (2,3,4 = x,y,c)
                                wm_setpixel(j,dstr[2],dstr[3],dstr[4]);
                                break;
                        case 1:         // Function 82/1: Get pixel (2,3 = x,y; ret: 4=c)
                                dstr[4] = wm_getpixel(j,dstr[2], dstr[3]);
                                break;
                        case 2:         // Function 82/2: Draw HLINE (2,3,4,5 = y,x1,x2,c)
                                wm_hline(j,dstr[2],dstr[3],dstr[4],dstr[5]);
                                break;
                        case 3:         // Function 82/3: Draw VLINE (2,3,4,5 = x,y1,y2,c)
                                wm_vline(j,dstr[2],dstr[3],dstr[4],dstr[5]);
                                break;
                        case 4:         // Function 82/4: Draw RECT (2,3,4,5,6 = x1,y1,x2,y2,c)
                                if(dstr[2] == 0xFFFF)
                                    wm_rect(j,0,0,curx-1,cury-1-tbheight-hbheight,dstr[6]);
                                else
                                    wm_rect(j,dstr[2],dstr[3],dstr[4],dstr[5],dstr[6]);
                                break;
                        case 5:         // Function 82/5: Draw FillRECT (2,3,4,5,6 = x1,y1,x2,y2,c)
                                if(dstr[2] == 0xFFFF)
                                    wm_rectf(j,0,0,curx-1,cury-1-tbheight-hbheight,dstr[6]);
                                else
                                    wm_rectf(j,dstr[2],dstr[3],dstr[4],dstr[5],dstr[6]);
                                break;
                        case 6:         // Function 82/6: Text (2,3,4,5,6,7,8 = x,y,fg,bg,font,seg,off)
                                if(dstr[3] == -1) dstr[3] = cury - (1+tbheight+hbheight + textheight(0,MK_FP(dstr[7],dstr[8])));
                                wm_outtext(j,dstr[2],dstr[3],dstr[4] == 0xFFFF ? 0xFFF : dstr[4], dstr[5] == 0xFFFF ? 0xFFF : dstr[5],0 /* ULDIF: Only font 0 is used so that ulsop-apps dont get confuest / dstr[6] */ ,MK_FP(dstr[7],dstr[8]));
                                printf("ulsop outtext done");
                                break;
                        case 7:         // Function 82/7: Draw sprite (2,3,4,5 = x,y,seg,off)
                                wm_drawsprite(j,dstr[2],dstr[3],MK_FP(dstr[4],dstr[5]));
                                break;
                        case 8:         // Function 82/8: Draw Button (2,3,4,5,6 = x,y,x2,y2,fillcol)
                                wm_drawbutton(j,dstr[2], dstr[3], dstr[4], dstr[5], dstr[6]);
                                break;
                        case 9:         // Function 82/9: Draw Buttondown (2,3,4,5,6 = x,y,x2,y2,fillcol)
                                wm_drawbuttondown(j,dstr[2], dstr[3], dstr[4], dstr[5], dstr[6]);
                                break;
                        case 0xA:       // Function 82/A: Draw Optionbutton (2,3,4 = X,Y,Down)
                                wm_drawoptbtn(j,dstr[2], dstr[3], dstr[4],0);
                                break;
                        case 0xB:       // Function 82/B: Draw Checkbox (2,3,4 = X,Y,Down)
                                wm_drawcheckbox(j,dstr[2], dstr[3], dstr[4],0);
                                break;
                        case 0xC:       // Function 82/C: Draw complete button (2,3,4,5,6,7,8,10,11,12 = X,Y,X2,Y2,Down,Font,FgCol,BgCol,Seg,Off)
                                wm_drawcompletebutton(j,dstr[2],dstr[3],dstr[4],dstr[5],dstr[6],dstr[7],dstr[8],dstr[10],MK_FP(dstr[11], dstr[12]));
                                break;
                        case 0xD:       // Function 82/D: Draw easybutton (2,3,4,5,6,7 = x,y,fntid,down,seg,off)
                                globfar = MK_FP(dstr[6],dstr[7]);
                                wm_drawcompletebutton(j,dstr[2],dstr[3],dstr[2] + textwidth(dstr[4],globfar) + 20,dstr[3] + textheight(dstr[4],globfar) + 5,dstr[5],dstr[4],0,7,globfar);
                                break;
                        case 0xE:       // Function 82/E: Paint area (2,3,4,5,6 = x,y,col,seg,off) seg:off=datastruct
                                // Datastruct (UINTS): xnum,ynum,leftspace,upspace,rightspace,downspace
                                globfar = MK_FP(dstr[5],dstr[6]);
                                globa = (curx - ((((UINT *) globfar)[2] + ((UINT *) globfar)[4]))) / ((UINT *) globfar)[0];
                                globb = (cury - ((tbheight+hbheight) + (((UINT *) globfar)[3] + ((UINT *) globfar)[5]))) / ((UINT *) globfar)[1];
                                globc = globa * dstr[2] + ((UINT *) globfar)[2];
                                globd = globb * dstr[3] + ((UINT *) globfar)[3];
                                wm_rectf(j,globc,globd,globc+globa-1,globd+globb-1,dstr[4]);
                                break;
                        case 0xF:       // Function 82/F: Draw rect around area (2,3,4,5,6 = x,y,col,seg,off) seg:off=datastruct
                                // Datastruct (UINTS): xnum,ynum,leftspace,upspace,rightspace,downspace,fillcol
                                globfar = MK_FP(dstr[5],dstr[6]);
                                globa = (curx - ((((UINT *) globfar)[2] + ((UINT *) globfar)[4]))) / ((UINT *) globfar)[0];
                                globb = (cury - ((hbheight+tbheight) + (((UINT *) globfar)[3] + ((UINT *) globfar)[5]))) / ((UINT *) globfar)[1];
                                globc = globa * dstr[2] + ((UINT *) globfar)[2];
                                globd = globb * dstr[3] + ((UINT *) globfar)[3];
                                wm_rectf(j,globc+1,globd+1,globc+globa-2,globd+globb-2,((UINT *) globfar)[6]);
                                wm_rect(j,globc,globd,globc+globa-1,globd+globb-1,dstr[4]);
                                break;
                        case 0x10:      // Function 82/10: Draw border around area (2 = color)
                                // Datastruct (UINTS): xnum,ynum,leftspace,upspace,rightspace,downspace
                /* Screen */    globa = (curx - ((((UINT *) globfar)[2] + ((UINT *) globfar)[4])));
                                globb = (cury - ((hbheight+tbheight) + (((UINT *) globfar)[3] + ((UINT *) globfar)[5])));
                  /* Tile */    globc = globa / ((UINT *) globfar)[0];
                                globd = globb / ((UINT *) globfar)[1];
                                globa = globc * ((UINT *) globfar)[0];
                                globb = globd * ((UINT *) globfar)[1];
                    /*  Left */ wm_rectf(j,((UINT *) globfar)[2],((UINT *) globfar)[3],((UINT *) globfar)[2] + globc - 1, globb - 1,dstr[2]);
                    /*    Up */ wm_rectf(j,((UINT *) globfar)[2],((UINT *) globfar)[3], globa - 1, globd - 1, dstr[2]);
                    /* Right */ wm_rectf(j,globa - globc, ((UINT *) globfar)[3], globa - 1, globb - 1, dstr[2]);
                    /*  Down */ wm_rectf(j,((UINT *) globfar)[2], globb - globd, globa - 1, globb - 1, dstr[2]);
                                break;
                        case 0x11:      // Function 82/11: Repaint GUI
                                break;
                        case 0x12:      // Function 82/12: Check for clicks (2 = x; 2=(ID for guiele)/(255=Nothing)/(x=Not fg.task)
                                break;
                        case 0x13:      // Function 82/13: Get 8-Bit Bitmap (2,3,4,5,6,7 = x,y,xsize,ysize,seg,off)
//                                getarea256(dstr[2],dstr[3],dstr[4]+dstr[2],dstr[5]+dstr[3],MK_FP(dstr[6],dstr[7]));
                                break;
                        case 0x14:      // Function 82/14: Put 8-Bit Bitmap (2,3,4,5,6,7 = x,y,xsize,ysize,seg,off)
//                                setarea256(dstr[2],dstr[3],dstr[4]+dstr[2],dstr[5]+dstr[3],MK_FP(dstr[6],dstr[7]));
                                break;
                        }
                break;
        }
    _asm
        {
        mov ax, rds
        mov ds, ax
        mov ax, res
        mov es, ax
        mov si, rsi
        mov di, rdi
        mov dx, rdx
        mov cx, rcx
        mov bx, rbx
        mov ax, rax
        mov sp, bp
        pop bp
        iret
        }
    }

ulsop_init()
    {
    printf("Initializing Ulsop-Compatibilitymodule.... ");

    setvect(0xFF, (void *) &ulhandle);       // Setting Servicehandler

    _asm
        {
        mov ax, 25FEh
        mov dx, OFFSET dstr
        int 21h                     // Pseudo-Int FEh Points to Param-Struct

            // Note: Ulsop-apps don't get into idle-mode (sleep,message,etc.).
            // They simple yield all there cpu-time if they don't need it
            // Ilas-apps will run fine next to ulsop-apps but the powermanagement
            // will be disabled if an ulsop-app is running (100% cpu usage)
;        mov ax, SEG idledata        // Ulsop idle = Ilas-idle (compatible)
;        mov ds, ax
        mov dx, OFFSET idledata
        mov ax, 25FCh
        int 21h

        mov ax, 25FDh               // Driver-chain (ilas uses a different technology, drivers are NOT available to ulvm)
        mov dx, OFFSET iretdata     // Pointer to an IRET (dummy)
        int 21h
        }
    printf("Done\n");
    }
